/**
 * \file
 * Functions and types for CRC checks.
 *
 * Generated on Sat Jun 15 14:34:05 2019
 * by pycrc v0.9.2, https://pycrc.org
 * using the configuration:
 *  - Width         = 16
 *  - Poly          = 0xed2f
 *  - XorIn         = 0xbeef
 *  - ReflectIn     = False
 *  - XorOut        = 0x0000
 *  - ReflectOut    = False
 *  - Algorithm     = table-driven
 *
 * This file defines the functions NVM_CRC_init(), NVM_CRC_update() and NVM_CRC_finalize().
 *
 * The NVM_CRC_init() function returns the inital \c crc value and must be called
 * before the first call to NVM_CRC_update().
 * Similarly, the NVM_CRC_finalize() function must be called after the last call
 * to NVM_CRC_update(), before the \c crc is being used.
 * is being used.
 *
 * The NVM_CRC_update() function can be called any number of times (including zero
 * times) in between the NVM_CRC_init() and NVM_CRC_finalize() calls.
 *
 * This pseudo-code shows an example usage of the API:
 * \code{.c}
 * NVM_CRC_t crc;
 * unsigned char data[MAX_DATA_LEN];
 * size_t data_len;
 *
 * crc = NVM_CRC_init();
 * while ((data_len = read_data(data, MAX_DATA_LEN)) > 0) {
 *     crc = NVM_CRC_update(crc, data, data_len);
 * }
 * crc = NVM_CRC_finalize(crc);
 * \endcode
 *
 * ## Additional Notes
 * 
 * The CRC polynomial (0xED2F) was chosen based on the research published by Philip Koopman of Carnegie Mellon
 * University [here](http://users.ece.cmu.edu/~koopman/crc/).  Dr. Koopman claims this polynomial has a
 * Hamming Distance of 10.
 *
 * The initial value, 0xBEEF, was chosen simply to avoid the most common EE values of 0xFFFF and 0x0000.
 * 
 */
#ifndef NVM_CRC_H
#define NVM_CRC_H

#include <stdlib.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif


/**
 * The definition of the used algorithm.
 *
 * This is not used anywhere in the generated code, but it may be used by the
 * application code to call algorithm-specific code, if desired.
 */
#define CRC_ALGO_TABLE_DRIVEN 1


/**
 * The type of the CRC values.
 *
 * CRCs are sixteen bits wide.
 */
typedef uint16_t NVM_CRC_t;


/**
 * Calculate the initial crc value.
 *
 * \return     The initial crc value.
 */
static inline NVM_CRC_t NVM_CRC_init(void)
{
    return 0xbeef;
}


/**
 * Update the crc value with new data.
 *
 * \param[in] crc      The current crc value.
 * \param[in] data     Pointer to a buffer of \a data_len bytes.
 * \param[in] data_len Number of bytes in the \a data buffer.
 * \return             The updated crc value.
 */
NVM_CRC_t NVM_CRC_update(NVM_CRC_t crc, const void *data, size_t data_len);


/**
 * Calculate the final crc value.
 *
 * \param[in] crc  The current crc value.
 * \return     The final crc value.
 */
static inline NVM_CRC_t NVM_CRC_finalize(NVM_CRC_t crc)
{
    return crc;
}


#ifdef __cplusplus
}           /* closing brace for extern "C" */
#endif

#endif      /* NVM_CRC_H */